@@ -2,19 +2,33 @@
 """Client for Hyperion servers."""
 
 import asyncio
+import collections
+import functools
 import inspect
 import json
 import logging
 import random
 import string
 import threading
-from typing import Any, Callable, Coroutine, Optional
+from typing import Any, Callable, Dict, Coroutine, Optional
 
 from hyperion import const
 
 _LOGGER = logging.getLogger(__name__)
 
 
+class HyperionError(Exception):
+    """Baseclass for all Hyperion exceptions."""
+
+
+class HyperionClientConnectAfterStartError(HyperionError):
+    """An exception indicating async_connect() called inappropriately."""
+
+
+class HyperionClientTanNotAvailable(HyperionError):
+    """An exception indicating the requested tan is not available."""
+
+
 class HyperionClient:
     """Hyperion Client."""
 
@@ -54,6 +68,11 @@ class HyperionClient:
         self._reader: Optional[asyncio.StreamReader] = None
         self._writer: Optional[asyncio.StreamWriter] = None
 
+        # Start tan @ 1, as the zeroth tan is used by default.
+        self._tan_cv = asyncio.Condition()
+        self._tan_counter = 1
+        self._tan_responses: Dict[int, Optional[Dict]] = collections.OrderedDict()
+
     def set_callbacks(self, callbacks: dict) -> None:
         """Set the update callbacks."""
         self._callbacks = callbacks
@@ -81,7 +100,17 @@ class HyperionClient:
         """Whether the client is actively managing the connection."""
         return self._manage_connection
 
-    async def async_connect(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_connect(self) -> bool:
+        """Connect to the Hyperion server after a safety check."""
+        if self._manage_connection_task is not None:
+            raise HyperionClientConnectAfterStartError(
+                "Cannot call async_connect() after starting background task "
+                "in Hyperion client (%s)",
+                self.id,
+            )
+        return await self._async_connect()
+
+    async def _async_connect(self) -> bool:
         """Connect to the Hyperion server."""
         future_streams = asyncio.open_connection(self._host, self._port)
         try:
@@ -102,7 +131,7 @@ class HyperionClient:
 
         # == Request: authorize ==
         if self._token is not None:
-            await self.async_login(token=self._token)
+            await self.async_send_login(token=self._token)
             resp_json = await self._async_safely_read_command()
             if (
                 not resp_json
@@ -120,7 +149,7 @@ class HyperionClient:
 
         # == Request: instance ==
         if self._instance != const.DEFAULT_INSTANCE:
-            await self.async_switch_instance(instance=self._instance)
+            await self.async_send_switch_instance(instance=self._instance)
             resp_json = await self._async_safely_read_command()
             if (
                 not resp_json
@@ -323,7 +352,7 @@ class HyperionClient:
     async def _async_manage_connection_once(self) -> None:
         """Manage the bidirectional connection to the server."""
         if not self._is_connected:
-            if not await self.async_connect():
+            if not await self._async_connect():
                 _LOGGER.info(
                     "Could not estalish valid connection to Hyperion (%s:%i), "
                     "retrying in %i seconds...",
@@ -344,7 +373,6 @@ class HyperionClient:
             _LOGGER.warning(
                 "Failed Hyperion (%s:%i) command: %s", self._host, self._port, resp_json
             )
-            return
         elif (
             command == f"{const.KEY_COMPONENTS}-{const.KEY_UPDATE}"
             and const.KEY_DATA in resp_json
@@ -422,6 +450,25 @@ class HyperionClient:
         elif command == f"{const.KEY_AUTHORIZE_LOGOUT}":
             await self.async_disconnect()
 
+        await self._handle_response_for_caller(command, resp_json)
+
+    async def _handle_response_for_caller(self, command, resp_json):
+        """Handle a server response for a caller."""
+
+        tan = resp_json.get(const.KEY_TAN)
+        if tan is not None:
+            async with self._tan_cv:
+                if tan in self._tan_responses:
+                    self._tan_responses[tan] = resp_json
+                    self._tan_cv.notify_all()
+                # Note: The behavior is not perfect here, in cases of an older
+                # Hyperion server and a malformed response. In that case, the server
+                # will return tan==0 (regardless of the input tan), and so the
+                # match here will fail. This will call the callee to time out
+                # awaiting a response (or wait forever if not timeout is
+                # specified). This was addressed in:
+                #
+                # https://github.com/hyperion-project/hyperion.ng/issues/1001 .
         self._call_callbacks(command, resp_json)
 
     # ==================
@@ -446,6 +493,73 @@ class HyperionClient:
         output.update(hard or {})
         return output
 
+    async def _reserve_tan_slot(self, tan: Optional[int] = None) -> int:
+        """Increment and return the next tan to use."""
+        async with self._tan_cv:
+            if tan is None:
+                # If tan is not specified, find the next available higher
+                # value.
+                while self._tan_counter in self._tan_responses:
+                    self._tan_counter += 1
+                tan = self._tan_counter
+                self._tan_counter += 1
+            if tan in self._tan_responses:
+                raise HyperionClientTanNotAvailable(
+                    "Requested tan '%i' is not available in Hyperion client (%s)"
+                    % (tan, self.id)
+                )
+            self._tan_responses[tan] = None
+            return tan
+
+    async def _remove_tan_slot(self, tan: int) -> None:
+        """Remove a tan slot that is no longer required."""
+        async with self._tan_cv:
+            if tan in self._tan_responses:
+                del self._tan_responses[tan]
+
+    async def _wait_for_tan_response(self, tan: int) -> Optional[Dict]:
+        """Wait for a response to arrive."""
+        async with self._tan_cv:
+            found = None
+            try:
+                found = await asyncio.wait_for(
+                    self._tan_cv.wait_for(
+                        lambda: self._tan_responses.get(tan) is not None
+                    ),
+                    timeout=self._timeout_secs,
+                )
+            except asyncio.TimeoutError:
+                pass
+
+            if not found:
+                return None
+            return self._tan_responses[tan]
+
+    class AwaitResponseWrapper:
+        """Wrapper an async *send* coroutine and await the response."""
+
+        def __init__(self, coro):
+            """Initialize the wrapper."""
+            self._coro = coro
+
+        async def __call__(self, client, *args, **kwargs):
+            """Call the wrapper."""
+            tan = await client._reserve_tan_slot(kwargs.get(const.KEY_TAN))
+            data = client._set_data(kwargs, hard={const.KEY_TAN: tan})
+
+            response = None
+            if await self._coro(client, *args, **data):
+                response = await client._wait_for_tan_response(tan)
+
+            await client._remove_tan_slot(tan)
+            return response
+
+        def __get__(self, instance, instancetype):
+            """Return a partial call that uses the correct 'self'."""
+            # Need to ensure __call__ receives the 'correct' outer
+            # 'self', which is 'instance' in this function.
+            return functools.partial(self.__call__, instance)
+
     # =============================
     # || Authorization API calls ||
     # =============================
@@ -455,7 +569,7 @@ class HyperionClient:
     # https://docs.hyperion-project.org/en/json/Authorization.html#authorization-check
     # ================================================================================
 
-    async def async_is_auth_required(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_is_auth_required(self, *args: Any, **kwargs: Any) -> bool:
         """Determine if authorization is required."""
         data = self._set_data(
             kwargs,
@@ -466,12 +580,14 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_is_auth_required = AwaitResponseWrapper(async_send_is_auth_required)
+
     # =============================================================================
     # ** Login **
     # https://docs.hyperion-project.org/en/json/Authorization.html#login-with-token
     # =============================================================================
 
-    async def async_login(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_login(self, *args: Any, **kwargs: Any) -> bool:
         """Login with token."""
         data = self._set_data(
             kwargs,
@@ -482,12 +598,14 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_login = AwaitResponseWrapper(async_send_login)
+
     # =============================================================================
     # ** Logout **
     # https://docs.hyperion-project.org/en/json/Authorization.html#logout
     # =============================================================================
 
-    async def async_logout(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_logout(self, *args: Any, **kwargs: Any) -> bool:
         """Logout."""
         data = self._set_data(
             kwargs,
@@ -498,12 +616,14 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_logout = AwaitResponseWrapper(async_send_logout)
+
     # ============================================================================
     # ** Request Token **
     # https://docs.hyperion-project.org/en/json/Authorization.html#request-a-token
     # ============================================================================
 
-    async def async_request_token(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_request_token(self, *args: Any, **kwargs: Any) -> bool:
         """Request an authorization token.
 
         The user will accept/deny the token request on the Web UI.
@@ -522,7 +642,9 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
-    async def async_request_token_abort(self, *args: Any, **kwargs: Any) -> bool:
+    async_request_token = AwaitResponseWrapper(async_send_request_token)
+
+    async def async_send_request_token_abort(self, *args: Any, **kwargs: Any) -> bool:
         """Abort a request for an authorization token."""
         data = self._set_data(
             kwargs,
@@ -534,6 +656,8 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_request_token_abort = AwaitResponseWrapper(async_send_request_token_abort)
+
     # ====================
     # || Data API calls ||
     # ====================
@@ -558,27 +682,31 @@ class HyperionClient:
             return
         self._serverinfo[const.KEY_ADJUSTMENT] = adjustment
 
-    async def async_set_adjustment(self, *args, **kwargs):
+    async def async_send_set_adjustment(self, *args, **kwargs):
         """Request that a color be set."""
         data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_ADJUSTMENT})
         return await self._async_send_json(data)
 
+    async_set_adjustment = AwaitResponseWrapper(async_send_set_adjustment)
+
     # =====================================================================
     # ** Clear **
     # Set: https://docs.hyperion-project.org/en/json/Control.html#clear
     # =====================================================================
 
-    async def async_clear(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_clear(self, *args: Any, **kwargs: Any) -> bool:
         """Request that a priority be cleared."""
         data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_CLEAR})
         return await self._async_send_json(data)
 
+    async_clear = AwaitResponseWrapper(async_send_clear)
+
     # =====================================================================
     # ** Color **
     # Set: https://docs.hyperion-project.org/en/json/Control.html#set-color
     # =====================================================================
 
-    async def async_set_color(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_color(self, *args: Any, **kwargs: Any) -> bool:
         """Request that a color be set."""
         data = self._set_data(
             kwargs,
@@ -587,6 +715,8 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_set_color = AwaitResponseWrapper(async_send_set_color)
+
     # ==================================================================================
     # ** Component **
     # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#components
@@ -621,13 +751,15 @@ class HyperionClient:
         else:
             new_components.append(new_component)
 
-    async def async_set_component(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_component(self, *args: Any, **kwargs: Any) -> bool:
         """Request that a color be set."""
         data = self._set_data(
             kwargs, hard={const.KEY_COMMAND: const.KEY_COMPONENTSTATE}
         )
         return await self._async_send_json(data)
 
+    async_set_component = AwaitResponseWrapper(async_send_set_component)
+
     def is_on(
         self,
         components: list = [const.KEY_COMPONENTID_ALL, const.KEY_COMPONENTID_LEDDEVICE],
@@ -670,7 +802,7 @@ class HyperionClient:
             return
         self._serverinfo[const.KEY_EFFECTS] = effects
 
-    async def async_set_effect(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_effect(self, *args: Any, **kwargs: Any) -> bool:
         """Request that an effect be set."""
         data = self._set_data(
             kwargs,
@@ -679,12 +811,14 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_set_effect = AwaitResponseWrapper(async_send_set_effect)
+
     # =================================================================================
     # ** Image **
     # Set: https://docs.hyperion-project.org/en/json/Control.html#set-image
     # =================================================================================
 
-    async def async_set_image(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_image(self, *args: Any, **kwargs: Any) -> bool:
         """Request that an image be set."""
         data = self._set_data(
             kwargs,
@@ -693,13 +827,15 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_set_image = AwaitResponseWrapper(async_send_set_image)
+
     # ================================================================================
     # ** Image Streaming **
     # Update: https://docs.hyperion-project.org/en/json/Control.html#live-image-stream
     # Set: https://docs.hyperion-project.org/en/json/Control.html#live-image-stream
     # ================================================================================
 
-    async def async_image_stream_start(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_image_stream_start(self, *args: Any, **kwargs: Any) -> bool:
         """Request a live image stream to start."""
         data = self._set_data(
             kwargs,
@@ -710,7 +846,9 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
-    async def async_image_stream_stop(self, *args: Any, **kwargs: Any) -> bool:
+    async_image_stream_start = AwaitResponseWrapper(async_send_image_stream_start)
+
+    async def async_send_image_stream_stop(self, *args: Any, **kwargs: Any) -> bool:
         """Request a live image stream to stop."""
         data = self._set_data(
             kwargs,
@@ -721,6 +859,8 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_image_stream_stop = AwaitResponseWrapper(async_send_image_stream_stop)
+
     # =================================================================================
     # ** Instances **
     # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#instance
@@ -739,7 +879,7 @@ class HyperionClient:
             return
         self._serverinfo[const.KEY_INSTANCE] = instances
 
-    async def async_start_instance(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_start_instance(self, *args: Any, **kwargs: Any) -> bool:
         """Start an instance."""
         data = self._set_data(
             kwargs,
@@ -750,7 +890,9 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
-    async def async_stop_instance(self, *args: Any, **kwargs: Any) -> bool:
+    async_start_instance = AwaitResponseWrapper(async_send_start_instance)
+
+    async def async_send_stop_instance(self, *args: Any, **kwargs: Any) -> bool:
         """Stop an instance."""
         data = self._set_data(
             kwargs,
@@ -761,7 +903,9 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
-    async def async_switch_instance(self, *args: Any, **kwargs: Any) -> bool:
+    async_stop_instance = AwaitResponseWrapper(async_send_stop_instance)
+
+    async def async_send_switch_instance(self, *args: Any, **kwargs: Any) -> bool:
         """Stop an instance."""
         data = self._set_data(
             kwargs,
@@ -772,6 +916,8 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_switch_instance = AwaitResponseWrapper(async_send_switch_instance)
+
     # =============================================================================
     # ** LEDs **
     # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#leds
@@ -807,18 +953,20 @@ class HyperionClient:
             return
         self._serverinfo[const.KEY_LED_MAPPING_TYPE] = led_mapping_type
 
-    async def async_set_led_mapping_type(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_led_mapping_type(self, *args: Any, **kwargs: Any) -> bool:
         """Request the LED mapping type be set."""
         data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_PROCESSING})
         return await self._async_send_json(data)
 
+    async_set_led_mapping_type = AwaitResponseWrapper(async_send_set_led_mapping_type)
+
     # ===================================================================================
     # ** Live LED Streaming **
     # Update: https://docs.hyperion-project.org/en/json/Control.html#live-led-color-stream
     # Set: https://docs.hyperion-project.org/en/json/Control.html#live-led-color-stream
     # ====================================================================================
 
-    async def async_led_stream_start(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_led_stream_start(self, *args: Any, **kwargs: Any) -> bool:
         """Request a live led stream to start."""
         data = self._set_data(
             kwargs,
@@ -829,7 +977,9 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
-    async def async_led_stream_stop(self, *args: Any, **kwargs: Any) -> bool:
+    async_led_stream_start = AwaitResponseWrapper(async_send_led_stream_start)
+
+    async def async_send_led_stream_stop(self, *args: Any, **kwargs: Any) -> bool:
         """Request a live led stream to stop."""
         data = self._set_data(
             kwargs,
@@ -840,6 +990,8 @@ class HyperionClient:
         )
         return await self._async_send_json(data)
 
+    async_led_stream_stop = AwaitResponseWrapper(async_send_led_stream_stop)
+
     # =================================================================================
     # ** Priorites **
     # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#priorities
@@ -887,11 +1039,13 @@ class HyperionClient:
             return
         self._serverinfo[const.KEY_PRIORITIES_AUTOSELECT] = priorities_autoselect
 
-    async def async_set_sourceselect(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_sourceselect(self, *args: Any, **kwargs: Any) -> bool:
         """Request the sourceselect be set."""
         data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_SOURCESELECT})
         return await self._async_send_json(data)
 
+    async_set_sourceselect = AwaitResponseWrapper(async_send_set_sourceselect)
+
     # ================================================================================
     # ** Sessions **
     # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#sessions
@@ -946,11 +1100,13 @@ class HyperionClient:
         if self._serverinfo:
             self._serverinfo[const.KEY_VIDEOMODE] = videomode
 
-    async def async_set_videomode(self, *args: Any, **kwargs: Any) -> bool:
+    async def async_send_set_videomode(self, *args: Any, **kwargs: Any) -> bool:
         """Request the LED mapping type be set."""
         data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_VIDEOMODE})
         return await self._async_send_json(data)
 
+    async_set_videomode = AwaitResponseWrapper(async_send_set_videomode)
+
 
 class ThreadedHyperionClient(HyperionClient, threading.Thread):
     """Hyperion Client that runs in a dedicated thread."""
@@ -985,7 +1141,7 @@ class ThreadedHyperionClient(HyperionClient, threading.Thread):
         )
 
         for name, value in inspect.getmembers(self):
-            if name.startswith("async_") and inspect.ismethod(value):
+            if name.startswith("async_") and callable(value):
                 new_name = name[len("async_") :]
                 self._register_sync_call(new_name, value)
 
@@ -999,17 +1155,34 @@ class ThreadedHyperionClient(HyperionClient, threading.Thread):
 
     def _async_wrapper(self, coro, *args: Any, **kwargs: Any) -> Any:
         """Convert a async call to synchronous by running it in the local event loop."""
-        task = coro(*args, **kwargs)
-        done, _ = self._loop.run_until_complete(asyncio.wait([task]))
-        if done:
-            return done.pop().result()
+        future = asyncio.run_coroutine_threadsafe(coro(*args, **kwargs), self._loop)
+        return future.result()
+
+    def stop(self):
+        """Stop the asyncio loop and thus the thread."""
+        self._loop.call_soon_threadsafe(self._loop.stop)
 
     def run(self) -> None:
-        """Run connection management in this thread."""
+        """Run the asyncio loop until stop is called."""
+        self._loop.run_forever()
 
-        async def manage(self):
-            while self._manage_connection:
-                await self._async_manage_connection_once()
 
-        self._manage_connection_task = self._loop.create_task(manage(self))
-        self._loop.run_until_complete(asyncio.wait([self._manage_connection_task]))
+class Response:
+    """Small wrapper class around a server response."""
+
+    def __init__(self, response):
+        """Initialize a Response object."""
+        self._response = response
+
+
+class ResponseOK(Response):
+    """Small wrapper class to determine if a response indicates success."""
+
+    @property
+    def is_success(self) -> bool:
+        """Determine if the response indicates success."""
+        return self._response and self._response.get(const.KEY_SUCCESS) is True
+
+    def __bool__(self):
+        """Determine if the response indicates success."""
+        return self.is_success
